#include <config.h>

module fix_gauge_class
  use comlib
  use lattice_class
  use error_class
  use timer_class
  use logfile_class
  use field_gauge_class
  use hmc_gluon_class, only : reunitarize
  implicit none
  private

  public :: fix_gauge
  public :: new,run,read,print,print_statistics,delete
  public :: fname_measure, read_config, save_config

  real(DP), parameter :: PLQ_TOL = 1.0e-12_DP

  type fix_gauge
  !
  ! job paramters for Ggauge Fixing
  ! Coulimb/Landau gauge fixing
  !
    character(len=CHARLEN) :: job_id       ! JOB ID
    character(len=CHARLEN) :: input_fname  ! input file name
    character(len=CHARLEN) :: cwd          ! current working dir.
    character(len=CHARLEN) :: fpath_in     !  input gauge file path
    character(len=CHARLEN) :: fhead_in     !  input gauge file name
    character(len=CHARLEN) :: fpath_out    ! output gauge file path
    character(len=CHARLEN) :: fhead_out    ! output gauge file name
    character(len=CHARLEN) :: reunit_log_fname = "gfix_reunit." ! reunitarization log file
    real(DP) :: eps1                       ! Stopping condition for derivative of gauge fixing functional.
    real(DP) :: eps2                       ! Stopping condition for relative diff. of gauge fixing functional.
    real(DP) :: omega                      ! Over Relaxation parameter for Mino's algorithm.
    real(DP) :: omega_new                  ! omega for retry.
    integer :: idim_max                    !  0: don't gauge fix, 3 : Coulomb gauge, 4 : Landau gauge.
    integer :: iter_max                    ! Max. number of iteration for gauge fixing
    integer :: isweep_CM                   ! Number of Cabibbo-Marinari sweep par site. (default=1)
    integer :: iter_OR_begin               ! Number of iteration to begin Mino's OR steps.
    integer :: iter_conv_check_begin       ! After iteration exceeds iter_check, convergence check begins.
    integer :: iter_print_skip             ! The iteration history is print out every iter_print_skip after iter_conv_check.
    integer :: iter_reunit_skip            ! every iter_reunit_skip, config. is reunitarized.
    integer :: itry_OR_max                 ! max number to retry Mino's OR.
    integer :: iter_inc_OR_begin           ! incremant of iter_OR at retry.
    integer :: itry_read                   ! 1: read configuration at retry, others: don't read.
    integer :: idummy1
    integer :: idummy2
    type(logfile) :: reunit_log
    type(timer) ::  tot_time
    type(timer) ::   io_time
    type(timer) :: gfix_time
  end type

  interface new
    module procedure new_gfix
  end interface

  interface delete
    module procedure delete_gfix
  end interface

  interface read
    module procedure read_gfix
  end interface

  interface print
    module procedure print_gfix
  end interface

  interface print_statistics
    module procedure print_stat_gfix
  end interface

  interface run
    module procedure run_gfix
  end interface

contains

function fname_measure(fpath,fhead) result(fname)
  implicit none
  character(CHARLEN) :: fname
  character(*), intent(in) :: fpath,fhead
  character(CHARLEN) :: fbuff,fdir
  integer :: i

  fname = REPEAT(' ',LEN(fname))
  fbuff = REPEAT(' ',LEN(fbuff))
  fdir = fpath
  i = LEN_TRIM(fdir)

  if (fdir(i:i) /= '/') then
    fdir = TRIM(ADJUSTL(fdir))//'/'
  endif

  write(fbuff,'(".x",z1,"y",z1,"z",z1)')(ipsite(i),i=1,NDIM-1)
  fname = TRIM(TRIM(ADJUSTL(fdir))//TRIM(ADJUSTL(fhead)))//TRIM(ADJUSTL(fbuff))
  
  return
end function

subroutine save_config(fpath,fhead,u,plq0)
  use file_tools_class
  use print_status_class
  implicit none
  character(len=CHARLEN), intent(in) :: fpath,fhead
  type(vfield_gluon_wg),  intent(in) :: u
  real(DP),               intent(in) :: plq0
  character(len=CHARLEN) :: fname, str
  real(DP) :: plq1
  integer :: iout

  fname = fname_measure(fpath,fhead)
  iout = search_free_file_unit()

  call plaquette(u,plq1)

  if (0 == nodeid) then
    write(*,'(80("-"))')
  endif
  write(str,'("      SAVE config : ",A)')TRIM(ADJUSTL(fname))
  call print_status(str)
  if (0 == nodeid) then
    write(*,'("    PLQ(Original) : ",E24.15)')plq0
    write(*,'("    PLQ(GFixed)   : ",E24.15)')plq1
  endif

  open(iout,file=TRIM(ADJUSTL(fname)),status='unknown',form='unformatted')
  write(iout)u
  write(iout)NTX,NTY,NTZ,NTT
  write(iout)NDIMX,NDIMY,NDIMZ
  write(iout)plq1
  close(iout)

  if (0 == nodeid) then
    write(*,'(80("-"))')
  endif

  return
end subroutine

subroutine read_config(fpath,fhead,u,reunit_log,plq0)
  use file_tools_class
  use print_status_class
  implicit none
  character(len=CHARLEN), intent(in)    :: fpath,fhead
  type(vfield_gluon_wg),  intent(inout) :: u
  type(logfile),          intent(inout) :: reunit_log
  real(DP),               intent(out)   :: plq0
  character(len=CHARLEN) :: fname, str
  real(DP) :: plq1,dq
  integer :: cNTX,cNTY,cNTZ,cNTT
  integer :: cNDIMX,cNDIMY,cNDIMZ
  integer :: iout

  fname = fname_measure(fpath,fhead)
  iout = search_free_file_unit()

  if (0 == nodeid) then
    write(*,'(80("-"))')
  endif
  write(str,'("      READ config : ",A)')TRIM(ADJUSTL(fname))
  call print_status(str)

  open(iout,file=TRIM(ADJUSTL(fname)),status='old',form='unformatted')
  read(iout)u
  read(iout)cNTX,cNTY,cNTZ,cNTT
  read(iout)cNDIMX,cNDIMY,cNDIMZ
  read(iout)plq0
  close(iout)

  call copy_boundary(u)
  call reunitarize(u,reunit_log)
  call copy_boundary(u)
  call plaquette(u,plq1)
  dq = ABS(plq1-plq0)/plq0

  if (0 == nodeid) then
    write(*,'("             PLQ0 : ",E24.15)')plq0
    write(*,'("             PLQ1 : ",E24.15)')plq1
    write(*,'(" |PLQ0-PLQ1|/PLQ0 : ",E24.15)')dq
    write(*,'(80("-"))')
  endif

  if (dq > PLQ_TOL) then
    write(str,'(" PLQ is differ ! STOP. (dq > PLQ_TOL =",E24.15,")")') PLQ_TOL
    call error_stop(TRIM(str))
  endif

  return
end subroutine


subroutine new_gfix(this)
  implicit none
  type(fix_gauge), intent(inout) :: this

  type(lattice_world) :: lattice
  character(len=8)  :: cdate
  character(len=10) :: ctime

  call new(lattice)

  call new(this%tot_time)
  call tic(this%tot_time)

  call new(this%gfix_time)
  call new(this%io_time)

  this%fpath_in   = repeat(' ',len(this%fpath_in))
  this%fhead_in   = repeat(' ',len(this%fhead_in))
  this%fpath_out  = repeat(' ',len(this%fpath_out))
  this%fhead_out  = repeat(' ',len(this%fhead_out))
  this%cwd        = repeat(' ',len(this%cwd))
  this%job_id     = repeat(' ',len(this%job_id))

  call DATE_AND_TIME(date=cdate,time=ctime)
  write(this%job_id,'(A,".",A)')TRIM(ADJUSTL(cdate)),TRIM(ADJUSTL(ctime))

  if (0 == nodeid) then
    call GET_COMMAND_ARGUMENT(1,this%input_fname)
    call getcwd(this%cwd)
  endif

  if (0 == nodeid) then
    write(*,'(" DATE AND TIME :",A)')TRIM(this%job_id)
  endif

#ifndef _singlePU
  if (NPU > 1) then
    call comlib_bcast(this%input_fname,0)
    call comlib_bcast(this%cwd,0)
    call comlib_bcast(this%job_id,0)
  endif
#endif

  this%reunit_log_fname=TRIM(ADJUSTL(this%reunit_log_fname))//TRIM(ADJUSTL(this%job_id))
  call new(this%reunit_log,TRIM(ADJUSTL(this%reunit_log_fname)))

  this%eps1 = EPSILON(1.0_DP)*2
  this%eps2 = EPSILON(1.0_DP)*2
  this%omega     = 1.99_DP
  this%omega_new = 1.90_DP

  this%idim_max  = 3
  this%iter_max  = 10000
  this%isweep_CM = 1
  this%iter_OR_begin = 500
  this%iter_conv_check_begin = 200
  this%iter_print_skip   = 100
  this%iter_reunit_skip  = 200
  this%itry_OR_max       = 10
  this%iter_inc_OR_begin = 50
  this%itry_read = 1

  return
end subroutine

subroutine delete_gfix(this)
  implicit none
  type(fix_gauge), intent(inout) :: this
  type(lattice_world) :: lattice
  call delete(this%reunit_log)
  call delete(lattice) 
  return
end subroutine

subroutine read_gfix(this)
!
! Read job parameters for Gauge Fixing (CM/OR method).
! The job parameters are read from input file with name "jobfile".
!
!  this : job parameters
!
  use file_tools_class
  implicit none
  type(fix_gauge), intent(inout) :: this
  integer :: iout

  call tic(this%io_time)

  iout = search_free_file_unit()
  if (0 == nodeid) then
    open(iout,file=this%input_fname,action='read',status='old',form='formatted')
    read(iout,*)
    read(iout,'(A)') this%fpath_in    ! input  config file path
    read(iout,'(A)') this%fhead_in    ! input  config file name head
    read(iout,*)
    read(iout,'(A)') this%fpath_out   ! output config file path
    read(iout,'(A)') this%fhead_out   ! output config file name head
    read(iout,*)
    read(iout,*) this%eps1
    read(iout,*) this%eps2
    read(iout,*) this%omega
    read(iout,*)
    read(iout,*) this%idim_max
    read(iout,*)
    read(iout,*) this%iter_max
    read(iout,*) this%isweep_CM
    read(iout,*) this%iter_OR_begin
    read(iout,*)
    read(iout,*) this%iter_conv_check_begin
    read(iout,*) this%iter_print_skip
    read(iout,*) this%iter_reunit_skip
    read(iout,*)
    read(iout,*) this%itry_OR_max
    read(iout,*) this%iter_inc_OR_begin
    read(iout,*) this%omega_new
    read(iout,*) this%itry_read
    close(iout)
  endif

  if (this%omega >= 2.0_DP) then
    this%omega=1.99_DP
    if (0 == nodeid) then
      write(*,'(" OR Omega should be < 2.")')
      write(*,'(" Reset,  omega = ",F24.16)') this%omega
    endif
  endif

  if (this%iter_reunit_skip <= 0) then
    this%iter_reunit_skip = 100 
    if (0 == nodeid) then
      write(*,'(" Re-Unitarization is preferred.")')
      write(*,'(" Reset,  iter_reunit_skip = ",I4)') this%iter_reunit_skip
    endif
  endif

#ifndef _singlePU
  if (NPU > 1) then
    call comlib_bcast(this%fpath_in,0)
    call comlib_bcast(this%fhead_in,0)
    call comlib_bcast(this%fpath_out,0)
    call comlib_bcast(this%fhead_out,0)
    call comlib_bcast(this%eps1,0)
    call comlib_bcast(this%eps2,0)
    call comlib_bcast(this%omega,0)
    call comlib_bcast(this%idim_max,0)
    call comlib_bcast(this%iter_max,0)
    call comlib_bcast(this%isweep_CM,0)
    call comlib_bcast(this%iter_OR_begin,0)
    call comlib_bcast(this%iter_conv_check_begin,0)
    call comlib_bcast(this%iter_print_skip,0)
    call comlib_bcast(this%iter_reunit_skip,0)
    call comlib_bcast(this%itry_OR_max,0)
    call comlib_bcast(this%iter_inc_OR_begin,0)
    call comlib_bcast(this%omega_new,0)
    call comlib_bcast(this%itry_read,0)
  endif
#endif

  call toc(this%io_time)

  return
end subroutine

subroutine print_config(u)
  implicit none
  type(vfield_gluon_wg), intent(inout) :: u
  integer :: itx,ity,itz,itt
  integer :: ipx,ipy,ipz,ipt
  integer :: idx,idy,idz,idt
  integer :: id,idd,ideo
  integer :: ibx,iby,ibz,ibt,ibth
  integer :: ieo,ieoxyz,ic,jc,mu

  if(0 == nodeid) then
    write(*,'(80("-"))')
    write(6,*) 'U(ic1,ic2;1,1,1,1,mu) ::'
    do mu=1,NDIM
      write(6,'("% mu =",I3)')mu
    do ic=1,COL
    do jc=1,COL
      write(6,'("%",2I4," (",ES22.14,",",ES22.14,")")')  &
 &                 ic,jc,u%eo(0)%mu(mu)%s(0,1,1,1)%u(ic,jc)
    enddo
    enddo
    enddo
    write(*,'(80("-"))')
  endif

#ifdef _DEBUG_
!  do itt=1,NTT
!  do itz=1,NTZ
!  do ity=1,NTY
!  do itx=1,NTX
  itt=1
  itz=1
  ity=1
  itx=1

    ieo  =mod(itt+itz+ity+itx,2)  ! global site even/odd index
!*** pu index 
    ipx=(itx-1)/NX
    ipy=(ity-1)/NY
    ipz=(itz-1)/NZ
    ipt=(itt-1)/NT
!*** block index in 1 pu
    idx =mod((itx-1)/NBX,NDX)
    idy =mod((ity-1)/NBY,NDY)
    idz =mod((itz-1)/NBZ,NDZ)
    idt =mod((itt-1)/NBT,NDT)
    ideo=mod((itt-1)/NBT+(itz-1)/NBZ+(ity-1)/NBY+(itx-1)/NBX,2)
    idd=idt+idz*NDT+idy*NDT*NDZ+idx*NDT*NDZ*NDY
    id=idd/2

!*** site index in a block
    ibx=mod(itx-1,NBX)+1
    iby=mod(ity-1,NBY)+1
    ibz=mod(itz-1,NBZ)+1
    ibt=mod(itt-1,NBT)+1
    ibth=ibt/2

    if (    (ipsite(1) == ipx) &
 &    .and. (ipsite(2) == ipy) &
 &    .and. (ipsite(3) == ipz) &
 &    .and. (ipsite(4) == ipt) ) then
      do mu=1,NDIM
        do ic=1,COL
        do jc=1,COL
          write(6,'(7I3,3X,2D20.12)') ic,jc,mu,itx,ity,itz,itt, &
 &              u%deo(ideo)%eo(ieo)%mu(mu)%b(id)%s(ibth,ibz,iby,ibx)%u(ic,jc)
        enddo
        enddo
      enddo
    endif
!  enddo
!  enddo
!  enddo
!  enddo
#endif

  if (0 == nodeid) then
    write(*,'(80("-"))')
  endif

  return
end subroutine

subroutine print_gfix(this)
!
! Print job parameters for Gauge Fixing (CM/OR method).
!
! this : job parameters
!
  implicit none
  type(fix_gauge), intent(inout) :: this

  if (0 == nodeid) then
    write(*,'(80("-"))')
    write(*,'("**** Lattice ****")')
    write(*,'("                Lattice Size :",4I3)')NTX,NTY,NTZ,NTT
    write(*,'("             PU Lattice Size :",4I3)')NX,NY,NZ,NT
    write(*,'("               PU Array Size :",4I3)')NDIMX,NDIMY,NDIMZ,1
    write(*,'("**** Gauge Fixing Parameters ****")')
    write(*,'("       Input conf. file path : ",A)')TRIM(this%fpath_in)
    write(*,'("       Input conf. file name : ",A)')TRIM(this%fhead_in)
    write(*,'("      Output conf. file path : ",A)')TRIM(this%fpath_out)
    write(*,'("      Output conf. file name : ",A)')TRIM(this%fhead_out)
    write(*,*)
    write(*,'("--------- Stopping Condition :")')
    write(*,'("- The program converges when both the following condition is satisfied.")')
    write(*,'("-    The successive change of the GF functional < EPS1.")')
    write(*,'("- The norm of the gradient of the GF functional < EPS2.")')
    write(*,'("                        EPS1 : ",E24.16)')this%eps1
    write(*,'("                        EPS2 : ",E24.16)')this%eps2
    write(*,*)
    write(*,'("---------- Fixing parameters :")')
    write(*,'("          OR Parameter Omega : ",E24.16)')this%omega
    write(*,'("                    IDIM_MAX : ",I3)')this%idim_max

    select case (this%idim_max)
    case (0)
      write(*,'("                             : => Do Not Fix Gauge.")')
    case (3)
      write(*,'("                             : => Couloumb Gauge Fixing.")')
    case (4)
      write(*,'("                             : =>   Landau Gauge Fixing.")')
    case default
      goto 100
    end select

    write(*,*)
    write(*,'("------- Iteration parameters :")')
    write(*,'("          Max # of Iteration : ",I10)')this%iter_max
    write(*,'(" # of Cabibbo-Marinari sweep : ",I10)')this%isweep_CM
    write(*,'("    # Iter. to start OR step : ",I10)')this%iter_OR_begin
    write(*,*)
    write(*,'("# Iter. to start Conv. Check : ",I10)')this%iter_conv_check_begin
    write(*,'("        # Skip Print History : ",I10)')this%iter_print_skip
    write(*,'("     # Skip Re-unitarization : ",I10)')this%iter_reunit_skip
    write(*,*)
    write(*,'("----------- Retry parameters :")')
    write(*,'("     Max. # of retry OR step : ",I10)')this%itry_OR_max
    write(*,'("# Iter to inclement OR start : ",I10)')this%iter_inc_OR_begin
    write(*,'("  New OR Parameter Omega_new : ",E24.16)')this%omega_new
    write(*,'(" Reload config. for retry ?  : ",I3)')this%itry_read

    select case (this%itry_read)
    case (1)
      write(*,'("                             : => Reload Config.")')
    case default
      write(*,'("                             : => Do Not Reload Config.")')
    end select

    write(*,'(80("-"))')
    write(*,*)
  endif

  return

100 continue
  call error_message(" Error in Input File, IDIM_MAX shoulr be 0, 3, or 4.")
  call error_stop(" Error Stop ! (read_param_gfix.F)")

end subroutine

subroutine print_stat_gfix(this)
!
! Print out job statistics
!
  implicit none
  type(fix_gauge), intent(inout) :: this
  real(DP) :: etime_gfix, etime_io, etime_copy, etime_tot

  etime_gfix = get_elapse(this%gfix_time)
  etime_io   = get_elapse(this%io_time)
  etime_copy = get_elapse(copy_sfg_time)

  call toc(this%tot_time)
  etime_tot = get_elapse(this%tot_time)

  if (0 == nodeid) then
    write(*,'(80("-"))')
    write(*,'("  Statistics par 1PU")')
    write(*,'("                 Total time (sec) :",F12.4)') etime_tot
    write(*,'("                    IO time (sec) :",F12.4," (",F8.4," %)")') etime_io  ,100*etime_io  /etime_tot
    write(*,'("          Gauge Fixing time (sec) :",F12.4," (",F8.4," %)")') etime_gfix,100*etime_gfix/etime_tot
    write(*,'("                copy_u time (sec) :",F12.4," (",F8.4," %)")') etime_copy,100*etime_copy/etime_tot
    write(*,'(80("="))')
  endif

  return
end subroutine

subroutine gfix_cmor(this,ifor,u)
!
! Gauge fixing with Cabibbo-Marinari / Mono's OR methods
!
! this : gauge fixing parameters
! ifor : 0 for CM, 1 for OR.
!    u : gauge field
!
  implicit none
  type(fix_gauge),       intent(inout) :: this
  integer,               intent(in)    :: ifor
  type(vfield_gluon_wg), intent(inout) :: u

  type(sfield_gluon_eo_wg), allocatable :: g
  integer :: ieo,ioe

  allocate(g)

  do ieo=0,1
    call new(g,ieo)
    ioe=1-ieo
    call gfix_cmor_eo(this,ifor,u%eo(ieo),u%eo(ioe),g)
    call copy_boundary(g)
    call gfix_trans_eo(this,u%eo(ieo),u%eo(ioe),g)
    call copy_boundary(u)
  enddo

  deallocate(g)

  return
end subroutine

subroutine gfix_cmor_eo(this,ifor,ue,uo,ge)
!
! Main routine for Coulomb/Landau Gauge Fixing.
! Do gauge transformation maximizing GF functional by
! Cabibbo-Marinari (CM) su(2) subgroup maximization method, and
! Mino's Over-Relaxation (OR) method.
!
!  this : input parameters
!  ifor : = 0: do CM method, 1 : do OR method.
!    ue : even/odd site gauge link
!    uo : odd/even site gauge link
!    ge : even/odd site gauge transfomation element
!
  implicit none
  type(fix_gauge),   intent(inout) :: this
  integer,           intent(in)    :: ifor
  type(vfield_gluon_eo_wg), intent(inout) :: ue,uo
  type(sfield_gluon_eo_wg), intent(inout) :: ge

  real(DP)    :: b0,b1,b2,b3,bab
  complex(DP) :: u11,u12
  complex(DP) :: uw(COL,COL)
  complex(DP) :: uz(COL,COL),dlt(COL,COL)
  complex(DP) :: zu1(COL),zu2(COL)
  complex(DP) :: zw1(COL),zw2(COL)

  complex(DP) :: ztmp
  integer :: i1,i2,iii,itest,iter2
  integer :: ix ,iy ,iz ,ith ,ieoxyz,ic,jc,mu
  integer :: ix2,iy2,iz2,ith2,ith0

  !--------------------------------------------
  ! su(2) subgroup of su(3) sweep ordering
  ! icol1: row index, icol2: colmun index
  !--------------------------------------------
  integer :: icol1(COL,COL*(COL-1))
  integer :: icol2(COL,COL*(COL-1))
  data  ((icol1(ic,itest),ic=1,COL),itest=1,COL*(COL-1) )  &
 &         / 1,1,2, 1,2,1, 1,1,2, 1,2,1, 2,1,1, 2,1,1 /
  data  ((icol2(ic,itest),ic=1,COL),itest=1,COL*(COL-1) )  &
 &         / 2,3,3, 2,3,3, 3,2,3, 3,3,2, 3,2,3, 3,3,2 /

  if ( (ue%ieo == uo%ieo) .or. (ge%ieo == uo%ieo) .or. (ge%ieo /= ue%ieo) ) then
    call error_stop("error in gfix_cmor_eo(site even/odd).")
  endif

  call new(ge,ue%ieo)

  !----------------------------------------------------
  !- itest=1,...6 : order to update SU(2) subgroups
  !      irandg = iand(nr*irandg,nzz)
  !       itest = int(azz6*irandg) + 1
  !----------------------------------------------------
  itest = 1

  select case (ifor)
  case (0)  ! Cabibbo-Marinari su(2) subgroup
!--------------------------------------------------------
! Cabibbo-Marinari su(2) subgroup sweep gauge fixing
!
! uw(x) = Sum_mu [U_mu(x)+U_mu^dag(x-mu)]
!
! gauge fixing function is derived from
!
! Sum_{x=odd or even only} Re[Tr[uw(x)]]
!  
!     = Sum_mu Sum_x Re[Tr[ U_mu(x) ]]
!--------------------------------------------------------
!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,ieoxyz,ith,ic,jc,mu,                   &
!$OMP         ix2,iy2,iz2,ith2,i1,i2,iii,iter2,               &
!$OMP         dlt,uw,b0,b1,b2,b3,bab,u11,u12,zu1,zu2,zw1,zw2)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz = mod(ipeo+ix+iy+iz+ue%ieo,2)
      do ith=1-ieoxyz,NTH-ieoxyz

        dlt(1,1) = Z1
        dlt(2,1) = Z0
        dlt(3,1) = Z0
        dlt(1,2) = Z0
        dlt(2,2) = Z1
        dlt(3,2) = Z0
        dlt(1,3) = Z0
        dlt(2,3) = Z0
        dlt(3,3) = Z1

!-------------------------------------------------------------------
! Gauge fixing local function
! uw(n) = Sum_mu[ u(n,mu) + u(n-mu,mu)^+]
!-------------------------------------------------------------------
        uw(:,:) = Z0
        do mu=1,this%idim_max
          ix2 =ix-isx(mu)  !  n + mu
          iy2 =iy-isy(mu)
          iz2 =iz-isz(mu)
          ith2=ith-(1-ieoxyz)*ist(mu)
          do jc=1,COL
          do ic=1,COL
            uw(ic,jc) = uw(ic,jc)                                      &
 &                   +       ue%mu(mu)%s(ith ,iz ,iy ,ix )%u(ic,jc)    &
 &                   + CONJG(uo%mu(mu)%s(ith2,iz2,iy2,ix2)%u(jc,ic))
          enddo
          enddo
        enddo

!-------------------------------------------------------------------
! SU(2) subgroup gauge fixing
!
!   Solutoin which maximize the local gauge fixng funciton 
! within su(2) subgroup.
!
!   | b0-ib3  -b2-ib1 |^+      | u11 u12 |
!   | b2-ib1   b0+ib3 |     =  | u21 u22 | = US,
!
!    u21=-CONJG(u12)
!    u22= CONJG(u11)
!
!-------------------------------------------------------------------
        do iter2=1,this%isweep_CM ! su(2) subgroup sweep iteration
        do iii=1,COL              ! 3 su(2) subgroup sweep 
          i1 = icol1(iii,itest)
          i2 = icol2(iii,itest)

          b0 =   REAL(uw(i1,i1)) + REAL(uw(i2,i2))
          b1 = -AIMAG(uw(i1,i2)) -AIMAG(uw(i2,i1))
          b2 =   REAL(uw(i2,i1)) - REAL(uw(i1,i2))
          b3 =  AIMAG(uw(i2,i2)) -AIMAG(uw(i1,i1))
          bab= b0**2+b1**2+b2**2+b3**2
          bab= 1.0_DP/SQRT(bab)
          u11 = bab * CMPLX(b0,b3,kind=DP)
          u12 = bab * CMPLX(b2,b1,kind=DP)

          !---------------------------------------------------
          ! Gauge transformation element using the solution.
          !
          !  dlt(n) <= US(n) dlt(n)
          !
          ! Update local gauge fixing function
          !
          !   uw(n) <= US(n) uw(n)
          !---------------------------------------------------
          zu1(1) =       u11  * dlt(i1,1)  &
 &                     + u12  * dlt(i2,1)
          zu1(2) =       u11  * dlt(i1,2)  &
 &                     + u12  * dlt(i2,2)
          zu1(3) =       u11  * dlt(i1,3)  &
 &                     + u12  * dlt(i2,3)

          zu2(1) = CONJG(u11) * dlt(i2,1)  &
 &                -CONJG(u12) * dlt(i1,1)
          zu2(2) = CONJG(u11) * dlt(i2,2)  &
 &                -CONJG(u12) * dlt(i1,2)
          zu2(3) = CONJG(u11) * dlt(i2,3)  &
 &                -CONJG(u12) * dlt(i1,3)

          zw1(1) =       u11  * uw(i1,1)  &
 &                     + u12  * uw(i2,1)
          zw1(2) =       u11  * uw(i1,2)  &
 &                     + u12  * uw(i2,2)
          zw1(3) =       u11  * uw(i1,3)  &
 &                     + u12  * uw(i2,3)

          zw2(1) = CONJG(u11) * uw(i2,1)  &
 &                -CONJG(u12) * uw(i1,1)
          zw2(2) = CONJG(u11) * uw(i2,2)  &
 &                -CONJG(u12) * uw(i1,2)
          zw2(3) = CONJG(u11) * uw(i2,3)  &
 &                -CONJG(u12) * uw(i1,3)

          dlt(i1,1) = zu1(1)
          dlt(i1,2) = zu1(2)
          dlt(i1,3) = zu1(3)
          dlt(i2,1) = zu2(1)
          dlt(i2,2) = zu2(2)
          dlt(i2,3) = zu2(3)
           uw(i1,1) = zw1(1)
           uw(i1,2) = zw1(2)
           uw(i1,3) = zw1(3)
           uw(i2,1) = zw2(1)
           uw(i2,2) = zw2(2)
           uw(i2,3) = zw2(3)

        enddo ! end of do iii   : 3 step in su(2) subgroup sweep
        enddo ! end of do iter2 : su(2) subgroup sweep iteration

        !-------------------------------------
        ! re-unitarize to make
        ! gauge transformation
        !
        ! re-unitarization a la Gram-Schmidt
        !-------------------------------------
        !--------------------------------
        ! normalize dlt(1)
        !--------------------------------
        b0 = REAL(dlt(1,1))**2+AIMAG(dlt(1,1))**2  &
 &         + REAL(dlt(2,1))**2+AIMAG(dlt(2,1))**2  &
 &         + REAL(dlt(3,1))**2+AIMAG(dlt(3,1))**2
        b0 = 1.0_DP/SQRT(b0)
        dlt(:,1) = b0 * dlt(:,1)

        !--------------------------------
        ! orthgonalize dlt(2) against
        !                        dlt(1)
        !--------------------------------
        ztmp = CONJG(dlt(1,1))*dlt(1,2)  &
 &           + CONJG(dlt(2,1))*dlt(2,2)  &
 &           + CONJG(dlt(3,1))*dlt(3,2)
        dlt(:,2) = dlt(:,2) - ztmp * dlt(:,1)

        b0 = REAL(dlt(1,2))**2+AIMAG(dlt(1,2))**2  &
 &         + REAL(dlt(2,2))**2+AIMAG(dlt(2,2))**2  &
 &         + REAL(dlt(3,2))**2+AIMAG(dlt(3,2))**2
        b0 = 1.0_DP/DSQRT(b0)
        dlt(:,2) = b0 * dlt(:,2)

        !--------------------------------
        ! dlt(3) = (dlt(1)xdlt(2))^*
        !  for SU(3) special unitarity
        !--------------------------------
        dlt(1,3) = CONJG(dlt(2,1)*dlt(3,2) - dlt(3,1)*dlt(2,2))
        dlt(2,3) = CONJG(dlt(3,1)*dlt(1,2) - dlt(1,1)*dlt(3,2))
        dlt(3,3) = CONJG(dlt(1,1)*dlt(2,2) - dlt(2,1)*dlt(1,2))

        !---------------------------------
        ! gauge transformation function
        !---------------------------------
        ge%s(ith ,iz ,iy ,ix )%u(:,:) = dlt(:,:)

      enddo
    enddo
    enddo
    enddo

  case (1) ! ifor == 1
!---------------------------------------------------------------
! Mino's OR gauge fixing
!---------------------------------------------------------------
!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,ieoxyz,ith,ic,jc,mu,         &
!$OMP         ix2,iy2,iz2,ith2,ztmp,                &
!$OMP         dlt,uw,uz,bab,b0,b1)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz = mod(ipeo+ix+iy+iz+ue%ieo,2)
      do ith=1-ieoxyz,NTH-ieoxyz

        !---------------------------------------------------------------
        ! uw(x) = Sum_mu [U_mu(x)-U_mu(x-mu)] : local gauge fixing function gradient
        ! uz(x) = Sum_mu [U_mu(x)+U_mu(x-mu)] : local gauge fixing function
        !
        ! Caution and do not confuse!
        ! uw is used for gauge fixing function in Cabibbo-Marinari step.
        !---------------------------------------------------------------
        uw(:,:) = Z0
        uz(:,:) = Z0
        do mu=1,this%idim_max
          ix2=ix-isx(mu)
          iy2=iy-isy(mu)
          iz2=iz-isz(mu)
          ith2=ith-(1-ieoxyz)*ist(mu)
          do jc=1,COL
          do ic=1,COL
            uw(ic,jc) = uw(ic,jc)  &
 &                         + ue%mu(mu)%s(ith ,iz ,iy ,ix )%u(ic,jc)  &
 &                         - uo%mu(mu)%s(ith2,iz2,iy2,ix2)%u(ic,jc)
            uz(ic,jc) = uz(ic,jc)  &
 &                         + ue%mu(mu)%s(ith ,iz ,iy ,ix )%u(ic,jc)  &
 &                         + uo%mu(mu)%s(ith2,iz2,iy2,ix2)%u(ic,jc)
          enddo
          enddo
        enddo ! end of do mu=1,this%idim_max

        !------------------------------------------------------
        ! calculate gradient of gauge fixing function
        !
        ! dlt <= - [ uw - uw^{dag} - (1/3) Tr[uw - uw^{dag}] ]
        !------------------------------------------------------
        bab = 2 * (AIMAG(uw(1,1))+AIMAG(uw(2,2))+AIMAG(uw(3,3)))/COL
        dlt(1,1) = CONJG(uw(1,1)) - uw(1,1) + CMPLX(0.0_DP,bab,kind=DP)
        dlt(2,1) = CONJG(uw(1,2)) - uw(2,1) 
        dlt(3,1) = CONJG(uw(1,3)) - uw(3,1) 
        dlt(1,2) = CONJG(uw(2,1)) - uw(1,2) 
        dlt(2,2) = CONJG(uw(2,2)) - uw(2,2) + CMPLX(0.0_DP,bab,kind=DP)
        dlt(3,2) = CONJG(uw(2,3)) - uw(3,2) 
        dlt(1,3) = CONJG(uw(3,1)) - uw(1,3) 
        dlt(2,3) = CONJG(uw(3,2)) - uw(2,3) 
        dlt(3,3) = CONJG(uw(3,3)) - uw(3,3) + CMPLX(0.0_DP,bab,kind=DP)

        !------------------------------------------------------
        ! Start Mino's OR gauge fixing
        ! Using the solution of gauge transformation
        ! equation to maximize gauge fixing functional.
        !------------------------------------------------------
        !-------------------------
        ! b0 <= Re[Tr[dlt * uw]]
        !-------------------------
        b0 =  &
 &         +  REAL(dlt(1,1))* REAL(uw(1,1))  &
 &         - AIMAG(dlt(1,1))*AIMAG(uw(1,1))  &
 &         +  REAL(dlt(1,2))* REAL(uw(2,1))  &
 &         - AIMAG(dlt(1,2))*AIMAG(uw(2,1))  &
 &         +  REAL(dlt(1,3))* REAL(uw(3,1))  &
 &         - AIMAG(dlt(1,3))*AIMAG(uw(3,1))  &
 &         +  REAL(dlt(2,1))* REAL(uw(1,2))  &
 &         - AIMAG(dlt(2,1))*AIMAG(uw(1,2))  &
 &         +  REAL(dlt(2,2))* REAL(uw(2,2))  &
 &         - AIMAG(dlt(2,2))*AIMAG(uw(2,2))  &
 &         +  REAL(dlt(2,3))* REAL(uw(3,2))  &
 &         - AIMAG(dlt(2,3))*AIMAG(uw(3,2))  &
 &         +  REAL(dlt(3,1))* REAL(uw(1,3))  &
 &         - AIMAG(dlt(3,1))*AIMAG(uw(1,3))  &
 &         +  REAL(dlt(3,2))* REAL(uw(2,3))  &
 &         - AIMAG(dlt(3,2))*AIMAG(uw(2,3))  &
 &         +  REAL(dlt(3,3))* REAL(uw(3,3))  &
 &         - AIMAG(dlt(3,3))*AIMAG(uw(3,3))

        !-------------------------
        ! uw <= dlt * dlt
        !-------------------------
        do jc=1,COL
        do ic=1,COL
          uw(ic,jc) = dlt(ic,1)*dlt(1,jc)  &
 &                  + dlt(ic,2)*dlt(2,jc)  &
 &                  + dlt(ic,3)*dlt(3,jc)
        enddo
        enddo

        !-------------------------
        ! b1 <= Re[Tr[uz * uw]]
        !-------------------------
        b1 =  &
 &         +  REAL(uz(1,1))* REAL(uw(1,1))  &
 &         - AIMAG(uz(1,1))*AIMAG(uw(1,1))  &
 &         +  REAL(uz(1,2))* REAL(uw(2,1))  &
 &         - AIMAG(uz(1,2))*AIMAG(uw(2,1))  &
 &         +  REAL(uz(1,3))* REAL(uw(3,1))  &
 &         - AIMAG(uz(1,3))*AIMAG(uw(3,1))  &
 &         +  REAL(uz(2,1))* REAL(uw(1,2))  &
 &         - AIMAG(uz(2,1))*AIMAG(uw(1,2))  &
 &         +  REAL(uz(2,2))* REAL(uw(2,2))  &
 &         - AIMAG(uz(2,2))*AIMAG(uw(2,2))  &
 &         +  REAL(uz(2,3))* REAL(uw(3,2))  &
 &         - AIMAG(uz(2,3))*AIMAG(uw(3,2))  &
 &         +  REAL(uz(3,1))* REAL(uw(1,3))  &
 &         - AIMAG(uz(3,1))*AIMAG(uw(1,3))  &
 &         +  REAL(uz(3,2))* REAL(uw(2,3))  &
 &         - AIMAG(uz(3,2))*AIMAG(uw(2,3))  &
 &         +  REAL(uz(3,3))* REAL(uw(3,3))  &
 &         - AIMAG(uz(3,3))*AIMAG(uw(3,3))

        !------------------------------------------------------
        ! solution of maximization equation of
        ! gauge fixing functional
        !
        ! b0 <= -omega * b0/b1 
        !     = -omega * Re[Tr[Dlt Uw]]/Re[Tr[ Dlt^2 Uz]]
        !     = -omega * Re[Tr[Dlt(x)  (U_mu(x)-U_mu(x-mu))]]
        !               /Re[Tr[Dlt(x)^2(U_mu(x)+U_mu(x-mu))]]
        !------------------------------------------------------
        b0 = -this%omega * b0 / b1

        !------------------------------------------------------
        ! calculate gauge transformation
        !
        !  dlt <= exp(b0 Dlt)
        !
        ! In this program, approximation
        !
        !  dlt <= (1 + b0 Dlt) \sim exp(b0 Dlt)
        ! 
        ! is used. And then dlt is re-unitarized.
        !------------------------------------------------------

        !---------------------------------
        ! upper 2 colmun is sufficient
        ! for re-unitalization.
        !---------------------------------
        dlt(1,1) = CMPLX(1.0_DP, b0*AIMAG(dlt(1,1)),kind=DP)
        dlt(2,1) =               b0*      dlt(2,1)
        dlt(3,1) =               b0*      dlt(3,1)
        dlt(1,2) =               b0*      dlt(1,2)
        dlt(2,2) = CMPLX(1.0_DP, b0*AIMAG(dlt(2,2)),kind=DP)
        dlt(3,2) =               b0*      dlt(3,2)

        !-------------------------------------
        ! re-unitarize to make
        ! gauge transformation
        !
        ! re-unitarization a la Gram-Schmidt
        !-------------------------------------
        !--------------------------------
        ! normalize dlt(1)
        !--------------------------------
        b0 = REAL(dlt(1,1))**2+AIMAG(dlt(1,1))**2  &
 &         + REAL(dlt(2,1))**2+AIMAG(dlt(2,1))**2  &
 &         + REAL(dlt(3,1))**2+AIMAG(dlt(3,1))**2
        b0 = 1.0_DP/SQRT(b0)
        dlt(:,1) = b0 * dlt(:,1)

        !--------------------------------
        ! orthgonalize dlt(2) against
        !                        dlt(1)
        !--------------------------------
        ztmp = CONJG(dlt(1,1))*dlt(1,2)  &
 &           + CONJG(dlt(2,1))*dlt(2,2)  &
 &           + CONJG(dlt(3,1))*dlt(3,2)
        dlt(:,2) = dlt(:,2) - ztmp * dlt(:,1)
        b0 = REAL(dlt(1,2))**2+AIMAG(dlt(1,2))**2  &
 &         + REAL(dlt(2,2))**2+AIMAG(dlt(2,2))**2  &
 &         + REAL(dlt(3,2))**2+AIMAG(dlt(3,2))**2
        b0 = 1.0_DP/DSQRT(b0)
        dlt(:,2) = b0 * dlt(:,2)

        !--------------------------------
        ! dlt(3) = (dlt(1)xdlt(2))^*
        !  for SU(3) special unitarity
        !--------------------------------
        dlt(1,3) = CONJG(dlt(2,1)*dlt(3,2) - dlt(3,1)*dlt(2,2))
        dlt(2,3) = CONJG(dlt(3,1)*dlt(1,2) - dlt(1,1)*dlt(3,2))
        dlt(3,3) = CONJG(dlt(1,1)*dlt(2,2) - dlt(2,1)*dlt(1,2))

        !---------------------------------
        ! gauge transformation function
        !---------------------------------
        ge%s(ith ,iz ,iy ,ix )%u(:,:) = dlt(:,:)

      enddo
    enddo
    enddo
    enddo ! end of do ix,iy,iz

  end select ! end of select case (ifor)

  return
end subroutine

subroutine gfix_trans_eo(this,ue,uo,ge)
!
! Do even/odd-site Gauge transformation
!
!  u(n   ,mu) <=  g(n) u(n,mu)
!  u(n-mu,mu) <=  u(n-mu,mu) g(n)^+
!
! this : gauge fixing parameters
!   ue : even/odd site gauge field
!   uo : odd/even site gauge field
!   ge : even/odd site gauge transformation field
!
  implicit none
  type(fix_gauge),          intent(inout) :: this
  type(vfield_gluon_eo_wg), intent(inout) :: ue,uo
  type(sfield_gluon_eo_wg), intent(inout) :: ge

  integer :: ieoxyz,mu,ic,jc
  integer :: ix ,iy ,iz ,ith
  integer :: ix2,iy2,iz2,ith2
  complex(DP) :: uge(COL,COL)

  if ( (ue%ieo == uo%ieo) .or. (ge%ieo == uo%ieo) .or. (ge%ieo /= ue%ieo) ) then
    call error_stop("Error in gfix_trans_eo (site even/odd)")
  endif

!------------------------------
!  u(x,mu) <=  g(x) u(x,mu)
!------------------------------
  do mu=1,NDIM
!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,ieoxyz,ith,ic,jc,uge)
  do ix=0,NX
  do iy=0,NY
  do iz=0,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ue%ieo,2)
    do ith=0,NTH-ieoxyz
      do jc=1,COL
      do ic=1,COL
        uge(ic,jc) = ge%s(ith ,iz ,iy ,ix )%u(ic, 1)  &
 &           *ue%mu(mu)%s(ith ,iz ,iy ,ix )%u( 1,jc)  &
 &                  +ge%s(ith ,iz ,iy ,ix )%u(ic, 2)  &
 &           *ue%mu(mu)%s(ith ,iz ,iy ,ix )%u( 2,jc)  &
 &                  +ge%s(ith ,iz ,iy ,ix )%u(ic, 3)  &
 &           *ue%mu(mu)%s(ith ,iz ,iy ,ix )%u( 3,jc)
      enddo
      enddo

      ue%mu(mu)%s(ith ,iz ,iy ,ix )%u(:,:) = uge(:,:)

    enddo
  enddo
  enddo
  enddo
  enddo

!---------------------------------------
!  u(x-mu,mu) <=  u(x-mu,mu) g(x)^+
!---------------------------------------
  do mu=1,NDIM
!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,ieoxyz,ith,ic,jc, &
!$OMP         ix2,iy2,iz2,ith2,uge)
  do ix=1,NX1
  do iy=1,NY1
  do iz=1,NZ1
    ieoxyz=mod(ipeo+ix+iy+iz+ue%ieo,2)
    do ith=1-ieoxyz,NTH
      ix2 = ix - isx(mu)
      iy2 = iy - isy(mu)
      iz2 = iz - isz(mu)
      ith2=ith-(1-ieoxyz)*ist(mu)

      do jc=1,COL
      do ic=1,COL
        uge(ic,jc) = uo%mu(mu)%s(ith2,iz2,iy2,ix2)%u(ic, 1)  &
 &                   *conjg(ge%s(ith ,iz ,iy ,ix )%u(jc, 1)) &
 &                  +uo%mu(mu)%s(ith2,iz2,iy2,ix2)%u(ic, 2)  &
 &                   *conjg(ge%s(ith ,iz ,iy ,ix )%u(jc, 2)) &
 &                  +uo%mu(mu)%s(ith2,iz2,iy2,ix2)%u(ic, 3)  &
 &                   *conjg(ge%s(ith ,iz ,iy ,ix )%u(jc, 3))
      enddo
      enddo

      uo%mu(mu)%s(ith2,iz2,iy2,ix2)%u(:,:) = uge(:,:)

    enddo
  enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine gfix_func(this,gfunc,u)
!
! Calculate the value of gauge fixing functional and 
! the norm of the gradient vector.
!
! gfunc(1) = Sum_{x}[ Sum_mu[Re[Tr[U(x,mu)]]     ]/COL/DIMMAX/Nvol
! gfunc(2) = Sum_{x}[ Re[Tr[Del(x)Del(x)^{dag}]] ]/COL/DIMMAX/Nvol
! gfunc(3) = Min_{x}[ Sum_mu[Re[Tr[U(x,mu)]]     ]/COL/DIMMAX
! gfunc(4) = Max_{x}[ Re[Tr[Del(x)Del(x)^{dag}]] ]/COL/DIMMAX
!
! where
!
!   Del(x) = V(x) - V(x)^{dag} - Tr[V(x) - V(x)^{dag}]/COL,
!
!  this : gauge fixing parameters
! gfunc : functional values
!     u : gauge field
!
  implicit none
  type(fix_gauge),       intent(in)  :: this
  real(DP),              intent(out) :: gfunc(4)
  type(vfield_gluon_wg), intent(in)  :: u

  real(DP) :: gfe(4),gfo(4)

  call gfix_func_eo(this,gfe,u%eo(0),u%eo(1))
  call gfix_func_eo(this,gfo,u%eo(1),u%eo(0))

  gfunc(1) =     gfe(1)+gfo(1)
  gfunc(2) =     gfe(2)+gfo(2)
  gfunc(3) = MIN(gfe(3),gfo(3))
  gfunc(4) = MAX(gfe(4),gfo(4))

  return
end subroutine

subroutine gfix_func_eo(this,gfunc,ue,uo)
!
!   this : gauge fixing parameters
!  gfunc : the value of gauge fixing functional and the norm of the gradient vector.
!     ue : even/odd site gauge link
!     uo : odd/even site gauge link
!
  implicit none
  type(fix_gauge),          intent(in)  :: this
  real(DP),                 intent(out) :: gfunc(4)
  type(vfield_gluon_eo_wg), intent(in)  :: ue,uo

  real(DP) :: gf1,gf2,gf3,gf4
  real(DP) :: rtmp,rsum(0:NPU-1)
  complex(DP) :: uv(COL,COL), del(COL,COL)
  integer :: ix ,iy ,iz ,ith ,ieoxyz,ic,jc,mu
  integer :: ix2,iy2,iz2,ith2


  if ( ue%ieo == uo%ieo ) then
    call error_stop("error in gfix_func_eo(site even/odd).")
  endif

  gf1 = 0.0_DP                            ! Sum_{x_even/odd}[Sum_mu[ Re[Tr[U(x,mu)]] ]]
  gf2 = 0.0_DP                            ! Sum_{x_even/odd}[ Re[Tr[Del(x)Del(x)^{dag}]] ]
  gf3 = REAL(this%idim_max*COL,kind=DP)   ! Min_{x_even/odd}[Sum_mu[ Re[Tr[U(x,mu)]] ]]
  gf4 = 0.0_DP                            ! Max_{x_even/odd}[ Re[Tr[Del(x)Del(x)^{dag}]] ]

!$OMP PARALLEL DO &
!$OMP PRIVATE(ix,iy,iz,ieoxyz,ith,ic,jc,mu,  &
!$OMP         ix2,iy2,iz2,ith2,uv,rtmp,del)  &
!$OMP REDUCTION(+:gf1,gf2) REDUCTION(MIN:gf3) REDUCTION(MAX:gf4)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz = mod(ipeo+ix+iy+iz+ue%ieo,2)
    do ith=1-ieoxyz,NTH-ieoxyz

      uv(:,:) = Z0
      rtmp = 0.0_DP
      do mu=1,this%idim_max

        !---------------------------------------
        ! gauge fixing functional 
        !      F[U]= Sum_{n,mu} Re[Tr[U(n,mu)]]
        !---------------------------------------
        rtmp = rtmp + REAL(ue%mu(mu)%s(ith,iz,iy,ix)%u(1,1))  &
 &                  + REAL(ue%mu(mu)%s(ith,iz,iy,ix)%u(2,2))  &
 &                  + REAL(ue%mu(mu)%s(ith,iz,iy,ix)%u(3,3))

        !---------------------------------------
        ! For gauge fixing functional gradient
        !      V(n) = Sum_mu[U(n,mu) - U(n-mu,mu)]
        !---------------------------------------
        ix2=ix-isx(mu)
        iy2=iy-isy(mu)
        iz2=iz-isz(mu)
        ith2=ith-(1-ieoxyz)*ist(mu)
        do jc=1,COL
        do ic=1,COL
          uv(ic,jc) = uv(ic,jc)  &
 &                  + ue%mu(mu)%s(ith ,iz ,iy ,ix )%u(ic,jc)  &
 &                  - uo%mu(mu)%s(ith2,iz2,iy2,ix2)%u(ic,jc)
        enddo
        enddo
      enddo ! end of do mu

      gf1  = gf1 + rtmp
      gf3  = MIN(gf3,rtmp)

      !------------------------------------------------
      ! Del(n) = V(n) - V^+(n)  - Tr[V(n) - V^+(n)]/3
      !------------------------------------------------
      rtmp = 2 * (AIMAG(uv(1,1))+AIMAG(uv(2,2))+AIMAG(uv(3,3)))/COL
    
      del(1,1) = uv(1,1) - CONJG(uv(1,1)) - CMPLX(0.0_DP,rtmp,kind=DP)
      del(2,1) = uv(2,1) - CONJG(uv(1,2))
      del(3,1) = uv(3,1) - CONJG(uv(1,3))
      del(1,2) = uv(1,2) - CONJG(uv(2,1))
      del(2,2) = uv(2,2) - CONJG(uv(2,2)) - CMPLX(0.0_DP,rtmp,kind=DP)
      del(3,2) = uv(3,2) - CONJG(uv(2,3))
      del(1,3) = uv(1,3) - CONJG(uv(3,1))
      del(2,3) = uv(2,3) - CONJG(uv(3,2))
      del(3,3) = uv(3,3) - CONJG(uv(3,3)) - CMPLX(0.0_DP,rtmp,kind=DP)
  
      !------------------------------------------------
      !  dF[U]/dG(n) \propt Tr[Del(n) Del(n)^+]
      !------------------------------------------------
      rtmp = REAL(del(1,1))**2 + AIMAG(del(1,1))**2  &
 &         + REAL(del(2,1))**2 + AIMAG(del(2,1))**2  &
 &         + REAL(del(3,1))**2 + AIMAG(del(3,1))**2  &
 &         + REAL(del(1,2))**2 + AIMAG(del(1,2))**2  &
 &         + REAL(del(2,2))**2 + AIMAG(del(2,2))**2  &
 &         + REAL(del(3,2))**2 + AIMAG(del(3,2))**2  &
 &         + REAL(del(1,3))**2 + AIMAG(del(1,3))**2  &
 &         + REAL(del(2,3))**2 + AIMAG(del(2,3))**2  &
 &         + REAL(del(3,3))**2 + AIMAG(del(3,3))**2

      gf2 = gf2 + rtmp
      gf4 = MAX(gf4,rtmp)

    enddo
  enddo
  enddo
  enddo

#ifndef _singlePU
  !-----------------------------
  ! global sum for gauge functional and the norm of the dradient vector
  !-----------------------------
  call comlib_sumcast(gf1)
  call comlib_sumcast(gf2)

  !-----------------------------
  ! min. value of local gauge fixing function
  !-----------------------------
  rsum(:) = 0.0_DP
  rsum(nodeid) = gf3
  call comlib_sumcast(rsum)
  gf3 = MINVAL(rsum)

  !-----------------------------
  ! max. value of local norm of the gradient vector
  !-----------------------------
  rsum(:) = 0.0_DP
  rsum(nodeid) = gf4
  call comlib_sumcast(rsum)
  gf4 = MAXVAL(rsum)
#endif

  rtmp = 1.0_DP/(this%idim_max*COL)          ! Normalization
  gfunc(1) = gf1*rtmp/REAL(NTX*NTY*NTZ*NTT,kind=DP)
  gfunc(2) = gf2*rtmp/REAL(NTX*NTY*NTZ*NTT,kind=DP)
  gfunc(3) = gf3*rtmp
  gfunc(4) = gf4*rtmp

  return
end subroutine

subroutine run_gfix(this)
!
!  Run gauge fixing
!
  implicit none
  type(fix_gauge), intent(inout) :: this

  type(vfield_gluon_wg), pointer  :: u
  integer :: ires,itry
  real(DP) :: plq,plq0

  allocate(u)
  call new(u)


  call tic(this%io_time)
  call read_config(this%fpath_in,this%fhead_in,u,this%reunit_log,plq0)
  call toc(this%io_time)

  call print_config(u)

  if (this%idim_max /= 0) then

    call gfix(this,u,plq,ires)

    if (this%iter_OR_begin < this%iter_max) then
      !--------------------------------------
      ! retry gauge fixing when normal
      ! gauge fixing fails to converge.
      !--------------------------------------
      itry = 0

      do while ((ires == 0) .AND. (itry < this%itry_OR_max))
        itry = itry + 1
        this%iter_OR_begin         = this%iter_OR_begin         + this%iter_inc_OR_begin
        this%iter_max              = this%iter_max              + this%iter_inc_OR_begin
        this%iter_conv_check_begin = this%iter_conv_check_begin + this%iter_inc_OR_begin
        this%omega                 = this%omega_new

        if (0 == nodeid) then
          write(6,*)
          write(6,'(" Normal Gauge Fixing is faild within the given job parameters.")')
          write(6,'(" Retry Gauge Fixing changing parameters.")')
          write(6,'("                  # of Retry : ",I9)')itry
          write(6,'("         Max. # of iteration : ",I9)') this%iter_max
          write(6,'("          Start iter. for OR : ",I9)') this%iter_OR_begin
          write(6,'(" Start iter. for Check Conv. : ",I9)') this%iter_conv_check_begin
          write(6,'("      New OR parameter Omega : ",E24.16)') this%omega
        endif

        if (this%itry_read == 1) then
          !---------------------------------------
          ! Reload gauge configuration for retry.
          !---------------------------------------
          call tic(this%io_time)
          call read_config(this%fpath_in,this%fhead_in,u,this%reunit_log,plq0)
          call toc(this%io_time)
        endif

        call gfix(this,u,plq,ires)

      enddo
    endif

  endif

  call tic(this%io_time)
  call print_config(u)

  !-----------------------------------
  ! save for measuremnt
  !-----------------------------------
  call save_config(this%fpath_out,this%fhead_out,u,plq0)
  call toc(this%io_time)

  deallocate(u)

  return
end subroutine

subroutine gfix(this,u,plq,iresult)
!
! Coulomb/Landau Gauge Fixing
!
!    this : input this parametes
!       u : gauge link
!     plq : plaquette value
! iresult : return gauge fixing status.  
!           1: converged, 0: fail to converge.
!
  implicit none
  type(fix_gauge),       intent(inout) :: this
  type(vfield_gluon_wg), intent(inout) :: u
  real(DP),              intent(out)   :: plq
  integer,               intent(out)   :: iresult

  integer, parameter :: Nmeas=4
  real(DP) :: epstmp1,epstmp2
  real(DP) :: gf0(Nmeas),gf1(Nmeas)
  logical :: is_reunitarized
  integer :: iter,ifor,i

  call tic(this%gfix_time)

  is_reunitarized = .true.

  if (0 == nodeid) then
    write(*,*)
    write(*,'(80("-"))')
    write(*,'("     <gf> := Sum_x[Sum_mu[Re[Tr[U(x,mu)]]]]/(COL*DIMMAX*Nvol)")')
    write(*,'("    <del> := Sum_x[Tr[Del(x)Del(x)^{dag}]] /(COL*DIMMAX*Nvol)")')
    write(*,'("  min(gf) := MIN[Sum_mu[Re[Tr[U(x,mu)]]]   /(COL*DIMMAX)")')
    write(*,'(" max(del) := MAX[Tr[Del(x)Del(x)^{dag}]]   /(COL*DIMMAX)")')
    write(*,'("  where")')
    write(*,'("    Del(x) = V(x)-V(x)^{dag} - Tr[V(x)-V(x)^{dag}]/COL,")')
    write(*,'("      V(x) = Sum_mu[U(x,mu) - U(x-mu,mu)].")')
    write(*,'("        mu = 1,2,..,DIMMAX.")')
    write(*,'(" rel_d<gf> : Relative change of <gf> in iteration.")')
    write(*,*)
    write(*,'(132("-"))')
    write(*,'("#",6X,"iter",10X,"<gf>",20X,"<del>",19X,"min(gf)",17X,"max(del)",14X,"rel_d<gf>")')
    write(*,'(132("-"))')
  endif

  call gfix_func(this,gf0,u) ! gauge functional and gradient.
  if (0 == nodeid) then
    write(*,'("@@",I9,10E24.16)') 0,(gf0(i),i=1,Nmeas),0.0_DP
  endif

  !-------------------------------
  ! start gauge fixing iteration
  !-------------------------------
  do iter=1,this%iter_max

!    write(*,fmt='(I9,",")',advance=no)iter
!    if (mod(iter,10)==0) write(*,*)
    !----------------------
    ! choose fixing method
    !----------------------
    if (iter < this%iter_OR_begin) then
      ifor = 0 ! Cabibbo-Marinari step
    else
      ifor = 1 ! Mono's OR step
    endif

    !-----------------------------------------------------------
    ! gauge fixing using Cabibbo-Marinari or Mino's OR methods.
    !-----------------------------------------------------------
    call gfix_cmor(this,ifor,u)
    is_reunitarized = .false.

    !-----------------------------------------------------------
    ! boundary copy and reunitarization
    !-----------------------------------------------------------
    if ( 0 == mod(iter,this%iter_reunit_skip) ) then
      call reunitarize(u,this%reunit_log)
      call copy_boundary(u)
      is_reunitarized = .true.
    endif

    !---------------------
    ! check convergince ?
    !---------------------
    if ( iter >= this%iter_conv_check_begin-1 ) then

      call gfix_func(this,gf1,u) ! gauge functional and gradient.
        
      epstmp1 = (gf1(1) - gf0(1))/gf1(1) ! The relative change of the GF functional (inclease)
      epstmp2 =  gf1(2)                  ! The norm of the gradient vector of GF functional

      if ( (iter >= this%iter_conv_check_begin) .and.  &
 &         (mod(iter,this%iter_print_skip) == 0) ) then 
        if (0 == nodeid) then
          write(*,'("@@",I9,10E24.16)') iter,(gf1(i),i=1,Nmeas),epstmp1
        endif
      endif
      if ( (epstmp1 >= 0.0_DP)    .and.  &
 &         (epstmp1 <  this%eps1) .and.  &
 &         (epstmp2 <  this%eps2) ) then
        if (0 == nodeid) then
          write(*,'("@@",I9,10E24.16)') iter,(gf1(i),i=1,Nmeas),epstmp1
        endif
        exit
      endif

      gf0(:)=gf1(:)  ! save the values for comparison in next iteration.

    endif

  enddo  ! end of do iter, main iteration loop

  !---------------------
  ! check convergence
  !---------------------
  if ( (this%iter_max >= this%iter_conv_check_begin) .and.  &
 &     (epstmp1 >= 0.0_DP)    .and.                         &
 &     (epstmp1 <  this%eps1) .and.                         &
 &     (epstmp2 <  this%eps2) ) then
    if (0 == nodeid) then
      write(*,'(132("-"))')
      write(*,'("@converged@ iter =",I9)') iter
    endif
    iresult = 1
  else 
    if (0 == nodeid) then
      write(*,'(132("-"))')
      write(*,'("@_NOT_converged@ iter =",I9)') iter-1
    endif
    iresult = 0
  endif

  !------------------------------
  ! reunitarize after gfixing
  !------------------------------
  if (.not.is_reunitarized) then
    call reunitarize(u,this%reunit_log)
    call copy_boundary(u)
  endif

  call plaquette(u,plq)

  if (0 == nodeid) then
    write(*,'(80("-"))')
    write(*,'("@gfunc@",10E24.16)')(gf1(i),i=1,Nmeas)
    write(*,'("@d<gf>@",E24.16)') epstmp1
    write(*,'("@<del>@",E24.16)') epstmp2
    write(*,'("@<PLQ>@",E24.16)') plq
    write(*,'(80("-"))')
    write(*,*)
  endif

  call toc(this%gfix_time)

  return
end subroutine

end module
