#include <config.h>
module timer_class
!
!= Defines \timer class
!
! provides \timer counter (counts on all nodes)
! uses CPU_TIME (fortran95 intrinsic)
! if you want more precise counter, change CPU_TIME.
!
!== Version
!
! $Id: timer_class.F90,v 1.20 2011/06/13 14:09:56 ishikawa Exp $
!
  use constants_module
  implicit none
  private
  public :: timer, new, delete, tic, toc
  public :: get_interval
  public :: get_elapse

  type timer
  !
  ! \timer object
  !
    private
    real(DP) ::    time0=0.0_DP  ! timer begin
    real(DP) ::    time1=0.0_DP  ! timer end
    real(DP) ::   elapse=0.0_DP  ! elapse time   (accumulates (time1-time0))
    real(DP) :: interval=0.0_DP  ! interval time (time1-time0)
    integer :: ticked = 0
    integer :: tocked = 0
    logical :: master_only = .false.
    logical :: bdummy = .false.
  end type

  interface new
    module procedure new_timer
  end interface

  interface delete
    module procedure delete_timer
  end interface

  interface get_interval
    module procedure get_interval_timer
  end interface

  interface get_elapse
    module procedure get_elapse_timer
  end interface

contains

subroutine new_timer(this,master_only)
!
! initialize \timer
!
  implicit none
  type(timer),       intent(inout) :: this
  logical, optional, intent(in)    :: master_only
  this%time0=0.0_DP
  this%time1=0.0_DP
  this%elapse=0.0_DP
  this%interval=0.0_DP
  this%ticked = 0
  this%tocked = 0
  if (present(master_only)) then
    this%master_only = master_only
  else
    this%master_only = .false.
  endif
  return
end subroutine

subroutine delete_timer(this)
!
! \delete \timer
!
  implicit none
  type(timer), intent(inout) :: this
  call new(this)
  return
end subroutine

subroutine tic(this)
!
! tick
!
#ifdef _OPENMP
  use omp_lib
#endif
  implicit none
  type(timer), intent(inout) :: this
  real(SP) :: time

#ifdef _OPENMP
  this%time0 = omp_get_wtime()
#else
  call CPU_TIME(time)
  this%time0 = real(time,kind=KIND(1.0_DP))
#endif
  this%ticked = 1
  this%tocked = 0

  return
end subroutine

subroutine toc(this)
!
! tock
!
#ifdef _OPENMP
  use omp_lib
#endif
  implicit none
  type(timer), intent(inout) :: this
  real(SP) :: time

#ifdef _OPENMP
  this%time1 = omp_get_wtime()
#else
  call CPU_TIME(time)
  this%time1 = real(time,kind=DP)
#endif
  this%interval = this%time1-this%time0
  if (this%ticked == 1) then
    this%elapse = this%elapse+this%time1-this%time0
  endif
  this%ticked = 0
  this%tocked = 1

  return
end subroutine

subroutine average_timer(this)
!
! averaging \timer count duaring all nodes
!
  use comlib
  use lattice_class, only : NPU
  implicit none
  type(timer), intent(inout) :: this

#ifndef _singlePU
  if (.not.this%master_only) then
    if (NPU > 1) then
      call comlib_sumcast(this%elapse)
      call comlib_sumcast(this%interval)
      this%elapse  =this%elapse/NPU
      this%interval=this%interval/NPU
    endif
  endif
#endif

  return
end subroutine

function get_interval_timer(this) result(interval)
!
! return interval time from first tick
!
  use comlib
  use lattice_class, only : NPU
  implicit none
  type(timer), intent(inout) :: this
  real(DP) :: interval
  interval = this%interval
#ifndef _singlePU
  if (.not.this%master_only) then
    if (NPU > 1) then
      call comlib_sumcast(interval)
    endif
  endif
#endif
  return
end function

function get_elapse_timer(this) result(elapse)
!
! return elapse time of the \timer
!
  use comlib
  implicit none
  type(timer), intent(inout) :: this
  real(DP) :: elapse
  call average_timer(this)
  elapse = this%elapse
  return
end function

end module
