#include <config.h>

module quark_dw_ovlp_class
!
!= Defines Domainwall fermion and truncated-DW-Overlap fermion
!
! Note chiral projection notation
!
! $ R y = (1-\gamma_5) y = P^{-} y $ , $ L y = (1+\gamma_5) y = P^{+} y $ 
!
! Domainwall operator notation:
!
!\[
!  (D_{DW}(m_q))(n,m)_{s,r}  =  (5-M) E(n,m) \otimes E_{s,r}
!                - \frac{1}{2} M_{hop}(n,m) \otimes E_{s,r} \] \[
!           - E(n,m) \otimes \frac{1}{2}
!                             \left[ (1-\gamma_5) \delta_{s+1,r}
!                                  + (1+\gamma_5) \delta_{s-1,r} \right.  \] \[ \left.
!                              - m_q (1-\gamma_5) \delta_{s,N}\delta_{1,r} 
!                              - m_q (1+\gamma_5) \delta_{s,1}\delta_{N,r} \right]
!\]
!
! Hoping operator notation:
!
!\[
!  M_{hop}(n,m) = \sum_{\mu=1}^{4}
!      \left[ (1-\gamma_{\mu})U_{\mu}(n)\delta_{n+\hat{\mu},m}
!           + (1+\gamma_{\mu})U_{\mu}^{\dag}(m)\delta_{n-\hat{\mu},m} \right]
!\]
!
!== Version
!
! $Id: quark_dw_ovlp_class.F90,v 1.22 2011/06/14 11:10:33 ishikawa Exp $
!
  use comlib
  use lattice_class
  use logfile_class
  use timer_class
  use error_class
  use solver_parameter_class
  use chrolog_class
!  use hmc_status_class
  use field_fermion_class
  use field_gauge_class
!  use quark_wilson_class
  use quark_dwf_class
  implicit none
  private

  public :: CRON_NSITE

  public :: new
  public :: delete

  public :: mult_iter
  public :: copy_fq_time

  public :: su3fv_spinor
  public :: field_quark_eo_wg
  public :: field_quark_eo_wog
  public :: field_quark_wg
  public :: field_quark_wog
  public :: set_gaussian_noise
  public :: mult_hopping_tzyx_eo 

  public :: dwf_low_modes
  public :: get_low_modes

  public :: field_dw_quark_wg
  public :: pack
  public :: unpack
  public :: abs2
  public :: prod
  public :: clear
  public :: assign
  public :: assign_add
  public :: accum_add
  public :: assign_sub
  public :: accum_sub
  public :: accum_mult
  public :: accum_mult_add
  public :: accum_add_mult
  public :: copy_boundary 
  public :: force_hmc_hopping

!  public :: quark_domainwall
  public :: set_id
  public :: get_id
  public :: get_NS
  public :: get_mass
  public :: set_coef
  public :: read
  public :: print
  public :: save_config
  public :: assign_mult_gamma5
  public ::  accum_mult_gamma5
  public :: assign_proj_5dto4d
  public :: assign_proj_4dto5d
  public :: assign_mult_impterm_dwf
  public :: assign_mult_dwf
  public :: assign_inv_mult_dwf

  public :: quark_truncd_dw_overlap
  public :: set_mass_physical
  public :: set_mass_pauli_villars
  public :: assign_mult_ov
  public :: assign_inv_mult_ov
  public :: assign_inv_mult_dagger_ov

  character(len=*), parameter :: ACTION_NAME = "Overlap Quark"

  type, extends(quark_domainwall) :: quark_truncd_dw_overlap
    real(DP), private ::   pv_mass = 1.0_DP   ! Pauli-Villars mass
    real(DP), private :: mass_buff = 1.0_DP   ! buffer for physical mass
  end type

  interface assign_mult_ov
    module procedure assign_mult_tdwov
    module procedure assign_mult_lowprojd_tdwov
  end interface

  interface assign_inv_mult_ov
    module procedure assign_inv_mult_tdwov
    module procedure assign_inv_mult_lowprojd_tdwov
  end interface

  interface assign_inv_mult_dagger_ov
    module procedure assign_inv_mult_dagger_tdwov
  end interface

contains

!--------------------------------------------------------------------
! Truncated-DW-Overlap type bound methods
!--------------------------------------------------------------------

subroutine set_mass_pauli_villars(this)
!
! set truncated dw overlap to Pauli-Villars mode
!
  implicit none
  type(quark_truncd_dw_overlap), intent(inout) :: this
  real(DP) :: mass
  this%mass_buff = get_mass(this)
  call set_mass(this,this%pv_mass)
  call make_mass_term(this)
  return
end subroutine

subroutine set_mass_physical(this)
!
! set truncated dw overlap to Physical mode.
!
  implicit none
  type(quark_truncd_dw_overlap), intent(inout) :: this
  call set_mass(this,this%mass_buff)
  this%mass_buff = get_mass(this)
!  write(*,*)"set_mass_physical"
  call make_mass_term(this)
  return
end subroutine

subroutine assign_mult_tdwov(this,iout,tol,iiter,dy,y,u)
!
! Assign and multiply truncated DW Overlap operator
!
! dy <= Dov y
!
!\[
!  D_{OV} = e^{\dag} P^{\dag} D_{DW}(1)^{-1} D_{DW}(m_{q}) P e
!\]
!
  implicit none
  type(quark_truncd_dw_overlap), intent(inout) :: this
  integer,  intent(in) :: iout
  real(DP), intent(in) :: tol
  integer,  intent(inout) :: iiter
  type(field_quark_wg), intent(inout) :: dy
  type(field_quark_wg), intent(inout) ::  y
  type(vfield_gluon_wg),intent(in)    ::  u
  type(field_dw_quark_wg), allocatable ::  py
  type(field_dw_quark_wg), allocatable ::  cy
  integer :: NS

!  write(*,*)"assign_mult_tdwov"

  allocate(py,cy)
  NS = get_NS(this)
  call new(cy,NS)
  call new(py,NS)

  call assign_proj_4dto5d(cy,y)

  !=========================================
  ! multiply Ddw(mq) : physical operator
  ! py = Ddw(mq) cy
  !=========================================
  call assign_mult_dwf(this,py,cy,u)

  !=========================================
  ! multiply inverse Ddw(1) : PV operator
  ! cy = Ddw(1) \ py
  !=========================================
  call set_mass_pauli_villars(this)
  call assign_inv_mult_dwf(this,iout,tol,iiter,cy,py,u)
  call set_mass_physical(this)
  call delete(py)

  call assign_proj_5dto4d(dy,cy)

  call delete(cy)
  deallocate(py,cy)

  return
end subroutine

subroutine assign_inv_mult_tdwov(this,iout,tol,iiter,dy,y,u,cron)
!
! Assign and multiply inverse truncated DW Overlap operator
!
! dy <= Dov \ y
!
!\[
! D_{OV} = e^{\dag} P^{\dag} D_{DW}(1)^{-1} D_{DW}(m_q) P e
!\]
!
!\[
! D_{OV}^{-1} = e^{\dag} P^{\dag} D_{DW}(m_q)^{-1} D_{DW}(1) P e
!\]
!
  implicit none
  type(quark_truncd_dw_overlap), intent(inout) :: this
  integer,  intent(in) :: iout
  real(DP), intent(in) :: tol
  integer,  intent(inout) :: iiter
  type(field_quark_wg), intent(inout) :: dy
  type(field_quark_wg), intent(inout) ::  y
  type(vfield_gluon_wg),intent(in)    ::  u
  type(chrolog_alg), optional, intent(inout) :: cron
  type(field_dw_quark_wg), allocatable ::  py
  type(field_dw_quark_wg), allocatable ::  cy
  integer :: NS

!  write(*,*)"assign_inv_mult_tdwov"

  allocate(py,cy)
  NS = get_NS(this)
  call new(cy,NS)
  call new(py,NS)

  call assign_proj_4dto5d(cy,y)

  !=========================================
  ! multiply Ddw(1) : PV operator
  ! py <= Ddw(1) cy
  !=========================================
  call set_mass_pauli_villars(this)
  call assign_mult_dwf(this,py,cy,u)
  call set_mass_physical(this)


  !==============================================
  ! multiply inverse Ddw(mq) : physical operator
  ! cy <= Ddw(mq) \ py
  !==============================================
  call assign_inv_mult_dwf(this,iout,tol,iiter,cy,py,u,cron=cron) 
  call delete(py)

  call assign_proj_5dto4d(dy,cy)
  call delete(cy)
  deallocate(py,cy)

  return
end subroutine

subroutine assign_inv_mult_dagger_tdwov(this,iout,tol,iiter,dy,y,u,cron)
!
! Assign and multiply inverse Hermitian conjugate of truncated DW Overlap operator
!
! dy <= Dov' \ y
!
!\[
! D_{OV}^{\dag} = e^{\dag} P^{\dag} D_{DW}(m_q)^{\dag} D_{DW}(1)^{\dag,-1} P e
!\]
!
!\[
! (D_{OV}^{\dag})^{-1} = e^{\dag} P^{\dag} D_{DW}(1)^{\dag} D_{DW}(m_q)^{\dag,-1} P e
!\]
!
  implicit none
  type(quark_truncd_dw_overlap), intent(inout) :: this
  integer,  intent(in) :: iout
  real(DP), intent(in) :: tol
  integer,  intent(inout) :: iiter
  type(field_quark_wg), intent(inout) :: dy
  type(field_quark_wg), intent(inout) ::  y
  type(vfield_gluon_wg),intent(in)    ::  u
  type(chrolog_alg), optional, intent(inout) :: cron

  type(field_dw_quark_wg), allocatable ::  py
  type(field_dw_quark_wg), allocatable ::  cy
  integer :: NS
  logical :: dagger

!  write(*,*)"assign_inv_mult_dagger_tdwov"

  dagger=.true.

  allocate(py,cy)
  NS = get_NS(this)
  call new(py,NS)
  call new(cy,NS)

  call assign_proj_4dto5d(py,y)

  !==============================================
  ! multiply inverse Ddw(mq)^dag : physical operator
  ! cy <= Ddw(mq)^dag \ py
  !==============================================
  call assign_inv_mult_dwf(this,iout,tol,iiter,cy,py,u,dagger=dagger,cron=cron) 

  !=========================================
  ! multiply Ddw(1)^dag : PV operator
  ! py <= Ddw(1)^dag cy
  !=========================================
  call set_mass_pauli_villars(this)
  call assign_mult_dwf(this,py,cy,u,dagger=dagger)
  call set_mass_physical(this)
  call delete(cy)

  call assign_proj_5dto4d(dy,py)
  call delete(py)
  deallocate(py,cy)

  return
end subroutine

subroutine assign_mult_lowprojd_tdwov(this,LV,iout,tol,iiter,dy,y,u)
!
! Assign and multiply truncated DW Overlap operator
! with kernel lowmode improvement
!
! dy <= Dov y
!
!\[
!  D_{OV} = e^{\dag} P^{\dag} D_{DW}(1)^{-1} D_{DW}(m_{q}) P e
!\]
!
  implicit none
  type(quark_truncd_dw_overlap), intent(inout) :: this
  type(dwf_low_modes),           intent(in)    :: LV
  integer,                       intent(in)    :: iout
  real(DP),                      intent(in)    :: tol
  integer,                       intent(inout) :: iiter
  type(field_quark_wg),          intent(inout) :: dy
  type(field_quark_wg),          intent(inout) ::  y
  type(vfield_gluon_wg),         intent(in)    ::  u
  type(field_dw_quark_wg), allocatable ::  py
  type(field_dw_quark_wg), allocatable ::  cy
  integer :: NS
  
  allocate(py,cy)
  NS = get_NS(this)
  call new(cy,NS)
  call new(py,NS)

  call assign_proj_4dto5d(cy,y)

  !=========================================
  ! multiply Ddw(mq) : physical operator
  ! py = Ddw(mq) cy
  !=========================================
  call assign_mult_dwf(this,py,cy,u,LV)

  !=========================================
  ! multiply inverse Ddw(1) : PV operator
  ! cy = Ddw(1) \ py
  !=========================================
  call set_mass_pauli_villars(this)
  call assign_inv_mult_dwf(this,iout,tol,iiter,cy,py,u,LV)

  call set_mass_physical(this)
  call delete(py)

  call assign_proj_5dto4d(dy,cy)

  call delete(cy)
  deallocate(py,cy)

  return
end subroutine

subroutine assign_inv_mult_lowprojd_tdwov(this,LV,iout,tol,iiter,dy,y,u)
!
! Assign and multiply inverse truncated DW Overlap operator
! with kernel lowmode improvement
!
! dy <= Dov \ y
!
!\[
! D_{OV} = e^{\dag} P^{\dag} D_{DW}(1)^{-1} D_{DW}(m_q) P e
!\]
!
!\[
! D_{OV}^{-1} = e^{\dag} P^{\dag} D_{DW}(m_q)^{-1} D_{DW}(1) P e
!\]
!
  implicit none
  type(quark_truncd_dw_overlap), intent(inout) :: this
  type(dwf_low_modes),           intent(in)    :: LV
  integer,                       intent(in)    :: iout
  real(DP),                      intent(in)    :: tol
  integer,                       intent(inout) :: iiter
  type(field_quark_wg),          intent(inout) :: dy
  type(field_quark_wg),          intent(inout) ::  y
  type(vfield_gluon_wg),         intent(in)    ::  u
  type(field_dw_quark_wg), allocatable ::  py
  type(field_dw_quark_wg), allocatable ::  cy
  integer :: NS

  allocate(py,cy)
  NS = get_NS(this)
  call new(cy,NS)
  call new(py,NS)

  call assign_proj_4dto5d(cy,y)

  !=========================================
  ! multiply Ddw(1) : PV operator
  ! py <= Ddw(1) cy
  !=========================================
  call set_mass_pauli_villars(this)
  call assign_mult_dwf(this,py,cy,u,LV)
  call set_mass_physical(this)

  !==============================================
  ! multiply inverse Ddw(mq) : physical operator
  ! cy <= Ddw(mq) \ py
  !==============================================
  call assign_inv_mult_dwf(this,iout,tol,iiter,cy,py,u,LV)
  call delete(py)

  call assign_proj_5dto4d(dy,cy)
  call delete(cy)
  deallocate(py,cy)

  return
end subroutine

end module
