#include <config.h>

module wavefunc_class
  use comlib
  use lattice_class
  implicit none
  private
  public :: wavefunc_param_obj
  public :: LOCAL_WAV
  public ::   EXP_WAV
  public ::  WALL_WAV
  public ::  POLY_WAV
  public :: new,delete,read,get,print

  integer, parameter :: MAXPOL=50
  integer, parameter :: LOCAL_WAV=1
  integer, parameter ::   EXP_WAV=2
  integer, parameter ::  WALL_WAV=3
  integer, parameter ::  POLY_WAV=4
  integer, parameter ::  USER_WAV=5

  integer, parameter :: NSPACE=NX*NY*NZ

  type wavefunc_param_obj
  ! quark source function parameters
    integer :: type                ! 1 for local, 2 for exponential, 3 for wall, 4 for polynomial, 5 user supplied
    integer :: itx0,ity0,itz0,itt0 ! source origin (1..NTX) (1..NTY) (1..NTZ) (1..NTT)
    integer :: Npolsf              ! number of polynomial coeffciient
    real(8) :: Asmear,Bsmear       ! exponetial parameter, psi(r)=A*exp(-r*B)
    real(8) :: Psmear(MAXPOL)      ! polynomial coeffcient, psi(r)= 1 - sum_(i=1,Npol) P(i)*r^i
    complex(8) :: wavefunc(COL,SPIN,NSPACE,NT) ! col,spin,zxy,t
  end type

  interface new
    module procedure new_wavefunc
  end interface

  interface read
    module procedure read_wavefunc
  end interface

  interface print
    module procedure print_wavefunc
  end interface

  interface get
    module procedure get_wavefunc
  end interface

  interface delete
    module procedure delete_wavefunc
  end interface

contains

function ispace(ix,iy,iz) result(isp)
  implicit none
  integer :: ix,iy,iz,isp
  isp=iz + (iy-1)*NZ + (ix-1)*NZ*NY
  return
end function

subroutine new_wavefunc(this)
  implicit none
  type(wavefunc_param_obj), intent(inout) :: this

  this%type = LOCAL_WAV
  this%itx0    = 1
  this%ity0    = 1
  this%itz0    = 1
  this%itt0    = 1
  this%wavefunc(:,:,:,:) = (0.0d0,0.0d0)

  return
end subroutine

subroutine delete_wavefunc(this)
  implicit none
  type(wavefunc_param_obj), intent(inout) :: this

  return
end subroutine

subroutine read_wavefunc(iout,this)
  implicit none
  integer, intent(in) :: iout
  type(wavefunc_param_obj), intent(inout) :: this
  integer :: ierr,i

  if (nodeid==0) then
    read(iout,*) this%type
    read(iout,*) this%itx0,this%ity0,this%itz0,this%itt0
  endif

  if (NPU>1) then
    call comlib_bcast(this%type,0)
    call comlib_bcast(this%itx0,0)
    call comlib_bcast(this%ity0,0)
    call comlib_bcast(this%itz0,0)
    call comlib_bcast(this%itt0,0)
  endif

  if ( this%type < 1 .and. this%type > 5) then
    write(*,'(" Wave function type error.")')
    write(*,'(" type = 1:loc, 2:exp, 3:wall, 4:pol, 5:usr")')
    stop
  endif
  if (   (this%itx0 < 1) .or. (this%itx0 > NTX)  &
 &  .or. (this%ity0 < 1) .or. (this%ity0 > NTY)  &
 &  .or. (this%itz0 < 1) .or. (this%itz0 > NTZ)  &
 &  .or. (this%itt0 < 1) .or. (this%itt0 > NTT) ) then
    write(*,'(" Source center (x,y,z,t) error.")')
    write(*,'(" x<=[1,NTX], y<=[1,NTY], z<=[1,NTZ], t<=[1,NTT]")')
    stop
  endif

  if (nodeid==0) then
    if (    ( this%type == LOCAL_WAV) .OR.  &
 &          ( this%type ==  WALL_WAV) ) then  ! local or wall source function
      continue
    else if ( this%type ==   EXP_WAV) then    ! exponential source function
      read(iout,*)this%Asmear,this%Bsmear 
    else if ( this%type ==   POLY_WAV) then   ! polynomial source function
      read(iout,*)this%Npolsf       ! read order of polynomial
      if ( this%Npolsf > MAXPOL) then
        write(*,'(" Error Npolsf > MAXPOL")')
        ierr=1
        goto 100
      endif
      read(iout,*)(this%Psmear(i),i=1,this%Npolsf) ! read polynomial coeffcients
    endif
  endif

100 if (ierr==1) stop

  if (NPU>1) then
    call comlib_bcast(this%Asmear,0)
    call comlib_bcast(this%Bsmear,0)
    call comlib_bcast(this%Npolsf,0)

    if (this%type== POLY_WAV) then
      do i=1,this%Npolsf
        call comlib_bcast(this%Psmear(i),0)
      enddo
    endif
  endif

  return
end subroutine

subroutine print_wavefunc(this)
  implicit none
  type(wavefunc_param_obj), intent(inout) :: this
  integer :: i

  if (nodeid==0) then
    write(*,'("==== Wave function Parameters ====")')
    write(*,'(" Wave function Center (xyzt):",4I3)') &
 &                 this%itx0,this%ity0,this%itz0,this%itt0 ! wavefunc center

    select case(this%type)
    case (LOCAL_WAV)
    write(*,'(3X," Local Source")')
    case (EXP_WAV)
    write(*,'(3X," Exponential Source")')
    write(*,'(3X,"          Asmear :",E24.16)')this%Asmear
    write(*,'(3X,"          Bsmear :",E24.16)')this%Bsmear
    case (WALL_WAV)
    write(*,'(3X," Wall Source")')
    case (POLY_WAV)
    write(*,'(3X," Polynomial Source")')
    write(*,'(3X,"           Order :",I3)')this%Npolsf
    write(*,'(3X,"       Psmear(:) :",100E24.16)') (this%Psmear(i),i=1,this%Npolsf)
    end select
    write(*,'(80("="))')
  endif

  return
end subroutine


subroutine get_wavefunc(this)
  implicit none
  type(wavefunc_param_obj), intent(inout) :: this
  complex(8), parameter :: z0=(0.0d0,0.0d0)
  complex(8), parameter :: z1=(1.0d0,0.0d0)
  integer :: iz,iy,ix,it,ic,is,jc,js,isp
  integer :: itt,itz,ity,itx
  integer :: rx,ry,rz,ipol
  real(8) :: rsrc,vsrc,rmax

  if ( this%type == USER_WAV) return

  this%wavefunc(:,:,:,:)=z0

  select case (this%type)
  case (LOCAL_WAV)

    do it=1,NT
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      itt=it+ipsite(4)*NT
      itz=iz+ipsite(3)*NZ
      ity=iy+ipsite(2)*NY
      itx=ix+ipsite(1)*NX
      if ( (itt==this%itt0) .and.  &
  &        (itz==this%itz0) .and.  &
  &        (ity==this%ity0) .and.  &
  &        (itx==this%itx0) ) then
        isp=ispace(ix,iy,iz)
        do is=1,SPIN
        do ic=1,COL
          this%wavefunc(ic,is,isp,it)=z1
        enddo
        enddo
      endif
    enddo
    enddo
    enddo
    enddo

  case (EXP_WAV)

    rmax = MIN(NTX,NTZ,NTY)
    rmax = (rmax/2.0d0 - 1.0d0) + 0.5d0
    
    do it=1,NT
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      itt=it+ipsite(4)*NT
      itz=iz+ipsite(3)*NZ
      ity=iy+ipsite(2)*NY
      itx=ix+ipsite(1)*NX

      isp=ispace(ix,iy,iz)

      if (itt==this%itt0) then
!        rx = itx - this%itx0
!        ry = ity - this%ity0
!        rz = itz - this%itz0
!
! 2-point distance with periodic boundary condition 
!
        rx=MIN(IABS(itx-this%itx0),IABS(itx-this%itx0+NTX),IABS(itx-this%itx0-NTX))
        ry=MIN(IABS(ity-this%ity0),IABS(ity-this%ity0+NTY),IABS(ity-this%ity0-NTY))
        rz=MIN(IABS(itz-this%itz0),IABS(itz-this%itz0+NTZ),IABS(itz-this%itz0-NTZ))
        rsrc = sqrt(real(rx**2+ry**2+rz**2,kind=KIND(rsrc)))
   
        if ( rsrc < rmax) then
          vsrc = this%Asmear*exp(-rsrc*this%Bsmear)
          do is=1,SPIN
          do ic=1,COL
            this%wavefunc(ic,is,isp,it)=cmplx(vsrc,0.0d0,kind=KIND(vsrc))
          enddo
          enddo
        endif

        if ( (rx==0) .and. (ry==0) .and. (rz==0) ) then
          do is=1,SPIN
          do ic=1,COL
            this%wavefunc(ic,is,isp,it)=z1
          enddo
          enddo
        endif

      endif

    enddo
    enddo
    enddo
    enddo

  case (WALL_WAV)

    vsrc = 1.0d0/real(NTX*NTY*NTZ,kind=KIND(vsrc))

    do it=1,NT
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      itt=it+ipsite(4)*NT
      itz=iz+ipsite(3)*NZ
      ity=iy+ipsite(2)*NY
      itx=ix+ipsite(1)*NX

      isp=ispace(ix,iy,iz)

      if (itt==this%itt0) then

        do is=1,SPIN
        do ic=1,COL
          this%wavefunc(ic,is,isp,it)=cmplx(vsrc,0.0d0,kind=KIND(vsrc))
        enddo
        enddo

      endif

    enddo
    enddo
    enddo
    enddo
  
  case (POLY_WAV)

    rmax = MIN(NTX,NTZ,NTY)
    rmax = (rmax/2.0d0 - 1.0d0) + 0.5d0
    
    do it=1,NT
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      itt=it+ipsite(4)*NT
      itz=iz+ipsite(3)*NZ
      ity=iy+ipsite(2)*NY
      itx=ix+ipsite(1)*NX

      isp=ispace(ix,iy,iz)

      if (itt==this%itt0) then
!        rx = itx - this%itx0
!        ry = ity - this%ity0
!        rz = itz - this%itz0
!
! 2-point distance with periodic boundary condition 
!
        rx=MIN(IABS(itx-this%itx0),IABS(itx-this%itx0+NTX),IABS(itx-this%itx0-NTX))
        ry=MIN(IABS(ity-this%ity0),IABS(ity-this%ity0+NTY),IABS(ity-this%ity0-NTY))
        rz=MIN(IABS(itz-this%itz0),IABS(itz-this%itz0+NTZ),IABS(itz-this%itz0-NTZ))
        rsrc = sqrt(real(rx**2+ry**2+rz**2,kind=KIND(rsrc)))
   
        vsrc = 1.0d0
        do ipol = 1,this%Npolsf
          vsrc = vsrc - this%Psmear(ipol)*rsrc**ipol
        enddo

        do is=1,SPIN
        do ic=1,COL
          this%wavefunc(ic,is,isp,it)=cmplx(vsrc,0.0d0,kind=KIND(vsrc))
        enddo
        enddo

      endif

    enddo
    enddo
    enddo
    enddo

  end select

  return
end subroutine

end module
